@@ -4,6 +4,7 @@ import re
 import json
 import fnmatch
 import traceback
+from typing import List, Tuple, Dict
 
 import speakeasy.common as common
 import speakeasy.winenv.arch as e_arch
@@ -18,6 +19,10 @@ EMU_ENGINES = (
               ('unicorn', unicorn_eng.EmuEngine),
               )
 
+WILDCARD_FLAG = bool
+API_LEVEL = Tuple[Dict[str, List[common.ApiHook]], WILDCARD_FLAG]
+MODULE_LEVEL = Tuple[Dict[str, API_LEVEL], WILDCARD_FLAG]
+
 
 # Generic emulator class for binary code
 class BinaryEmulator(MemoryManager):
@@ -96,6 +101,7 @@ class BinaryEmulator(MemoryManager):
         self.osversion = config.get('os_ver', {})
         self.env = config.get('env', {})
         self.user_config = config.get('user', {})
+        self.domain = config.get('domain')
         self.hostname = config.get('hostname')
         self.symlinks = config.get('symlinks', [])
         self.config_modules = config.get('modules', {})
@@ -108,6 +114,7 @@ class BinaryEmulator(MemoryManager):
         self.timeout = config.get('timeout', 0)
         self.max_api_count = config.get('max_api_count', 5000)
         self.exceptions = config.get('exceptions', {})
+        self.drive_config = config.get('drives', [])
         self.filesystem_config = config.get('filesystem', {})
         self.keep_memory_on_free = config.get('keep_memory_on_free', False)
 
@@ -139,6 +146,12 @@ class BinaryEmulator(MemoryManager):
                 verstr = '%s.%d_%d' % (os_name, major, minor)
                 return verstr
 
+    def get_domain(self):
+        """
+        Get domain of the machine being emulated
+        """
+        return self.domain
+
     def get_hostname(self):
         """
         Get hostname of the machine being emulated
@@ -198,6 +211,12 @@ class BinaryEmulator(MemoryManager):
         """
         return self.filesystem_config
 
+    def get_drive_config(self):
+        """
+        Get the drive settings specified in the drives section of the config file
+        """
+        return self.drive_config
+
     def reg_write(self, reg, val):
         """
         Write a value to an emulated cpu register
@@ -241,20 +260,18 @@ class BinaryEmulator(MemoryManager):
                 tu = [i for i in self.disasm_eng.disasm_lite(bytes(mem), addr)]
                 address, size, mnem, oper = tu[0]
             else:
-                tu = [(i.mnemonic, i.op_str) for i in
-                      self.disasm_eng.disasm(bytes(mem), addr)]
-                mnem, oper = tu[0]
+                return [i for i in self.disasm_eng.disasm(bytes(mem), addr)]
         except IndexError:
             raise EmuException("Failed to disasm at address: 0x%x" % (addr))
 
         op = '%s %s' % (mnem, oper)
         return ((mnem, oper, op))
 
-    def disasm(self, mem, addr):
+    def disasm(self, mem, addr, fast=True):
         """
         Disassemble bytes at a specified address
         """
-        return self._cs_disasm(mem, addr)
+        return self._cs_disasm(mem, addr, fast=fast)
 
     def get_register_state(self):
         """
@@ -279,23 +296,27 @@ class BinaryEmulator(MemoryManager):
                               ('rip', e_arch.AMD64_REG_RIP),
                               ('rsi', e_arch.AMD64_REG_RSI),
                               ('rdi', e_arch.AMD64_REG_RDI),
+                              ('rax', e_arch.AMD64_REG_RAX),
+                              ('rbx', e_arch.AMD64_REG_RBX),
+                              ('rcx', e_arch.AMD64_REG_RCX),
+                              ('rdx', e_arch.AMD64_REG_RDX),
                               ('r8',  e_arch.AMD64_REG_R8),
                               ('r9',  e_arch.AMD64_REG_R9),
                               ('r10', e_arch.AMD64_REG_R10),
                               ('r11', e_arch.AMD64_REG_R11),
-                              ('rax', e_arch.AMD64_REG_RAX),
-                              ('rbx', e_arch.AMD64_REG_RBX),
-                              ('rcx', e_arch.AMD64_REG_RCX),
-                              ('rdx', e_arch.AMD64_REG_RDX)):
+                              ('r12', e_arch.AMD64_REG_R12),
+                              ('r13', e_arch.AMD64_REG_R13),
+                              ('r14', e_arch.AMD64_REG_R14),
+                              ('r15', e_arch.AMD64_REG_R15)):
                 val = self.reg_read(reg)
                 regs[name] = "{0:#0{1}x}".format(val, 2 + (self.get_ptr_size() * 2))
         return regs
 
-    def get_disasm(self, addr, size):
+    def get_disasm(self, addr, size, fast=True):
         """
         Get the disassembly from an address
         """
-        return self.disasm(self.mem_read(addr, size), addr)
+        return self.disasm(self.mem_read(addr, size), addr, fast)
 
     def set_func_args(self, stack_addr, ret_addr, *args, home_space=True):
         """
@@ -393,7 +414,7 @@ class BinaryEmulator(MemoryManager):
 
         return argv
 
-    def do_call_return(self, ret_addr, argc, ret_value=None, conv=None):
+    def do_call_return(self, argc, ret_addr=None, ret_value=None, conv=None):
         """
         Set the emulation state after a call has completed
         """
@@ -413,8 +434,9 @@ class BinaryEmulator(MemoryManager):
 
         stk_ptr = self.reg_read(sp)
 
-        self.reg_write(sp, stk_ptr + self.ptr_size)
-        self.reg_write(pc, ret_addr)
+        if ret_addr:
+            self.reg_write(sp, stk_ptr + self.ptr_size)
+            self.reg_write(pc, ret_addr)
         if ret_value is not None:
             self.reg_write(rr, ret_value)
 
@@ -484,17 +506,32 @@ class BinaryEmulator(MemoryManager):
         elif self.get_arch() == e_arch.ARCH_AMD64:
             self.reg_write(e_arch.AMD64_REG_RSP, addr)
 
+    def format_stack(self, num_ptrs):
+        """
+        Get the stack and format it for display
+        """
+        out = []
+        sp = self.get_stack_ptr()
+        for i in range(num_ptrs):
+            try:
+                ptr = self.mem_read(sp, self.get_ptr_size())
+            except Exception:
+                return out
+            ptr = int.from_bytes(ptr, 'little')
+            tag = self.get_address_tag(ptr)
+            out.append((sp, ptr, tag))
+            sp += self.get_ptr_size()
+        return out
+
     def print_stack(self, num_ptrs):
         """
         This a debug function used to print the current stack state
         """
-        sp = self.get_stack_ptr()
+        ptrs = self.format_stack(num_ptrs)
         print('Stack:')
         print('***********************')
-        for i in range(num_ptrs):
-            ptr = self.mem_read(sp, self.get_ptr_size())
-            ptr = int.from_bytes(ptr, 'little')
-            tag = self.get_address_tag(ptr)
+        for p in ptrs:
+            sp, ptr, tag = p
             if tag:
                 fmt = 'sp=0x%x:\t0x%x\t->\t%s' % (sp, ptr, tag)
             else:
@@ -649,7 +686,7 @@ class BinaryEmulator(MemoryManager):
         i = 0
 
         if width == 1:
-            decode = 'utf-8'
+            decode = 'latin1'
         elif width == 2:
             decode = 'utf-16le'
         else:
@@ -691,10 +728,13 @@ class BinaryEmulator(MemoryManager):
         pat = b'[\x20-\x7f]{%d,}' % (min_len)
         res = re.compile(pat)
         hits = res.findall(data)
+        offset = 0
         for s in hits:
             try:
+                offset = data.find(s, offset)
                 s = s.decode('utf-8')
-                astrs.append(s)
+                astrs.append((offset, s))
+                offset += 1
             except UnicodeDecodeError:
                 continue
         return astrs
@@ -707,10 +747,13 @@ class BinaryEmulator(MemoryManager):
         pat = b'(?:[\x20-\x7f]\x00){%d,}' % (min_len)
         res = re.compile(pat)
         hits = res.findall(data)
+        offset = 0
         for ws in hits:
             try:
+                offset = data.find(ws, offset)
                 ws = ws.decode('utf-16le')
-                wstrs.append(ws)
+                wstrs.append((offset, ws))
+                offset += 1
             except UnicodeDecodeError:
                 continue
         return wstrs
@@ -787,28 +830,81 @@ class BinaryEmulator(MemoryManager):
                 return mod[0]
         return None
 
-    def get_api_hook(self, mod_name, func_name):
+    def get_api_hooks(self, mod_name, func_name) -> List[common.ApiHook]:
         """
         If an API hook has been set, return it here
         """
-        for hook in self.hooks.get(common.HOOK_API, []):
-            if fnmatch.fnmatch(mod_name.lower(), hook.module.lower()):
-                if fnmatch.fnmatch(func_name.lower(), hook.api_name.lower()):
-                    return hook
-        return None
 
-    def add_api_hook(self, cb, module='', api_name='', argc=0, call_conv=None, emu=None):
+        mod_name = mod_name.lower()
+        func_name = func_name.lower()
+        try:
+            hook_struct, wildcard_module = self.hooks[common.HOOK_API]
+        except KeyError:
+            return []
+        try:
+            modules = [hook_struct[mod_name]]
+        except KeyError:
+            modules = []
+        if wildcard_module:
+            for module_name_saved, value in hook_struct.items():
+                if fnmatch.fnmatch(mod_name, module_name_saved) and mod_name != module_name_saved:
+                    modules.append(value)
+        user_hooks = []
+        for module in modules:
+            hooks, wildcard_api = module
+            try:
+                user_hooks.extend(hooks[func_name])
+            except KeyError:
+                pass
+            if wildcard_api:
+                for func_name_saved, list_of_hooks in hooks.items():
+                    if (fnmatch.fnmatch(func_name, func_name_saved) and
+                       func_name != func_name_saved):
+                        user_hooks.extend(list_of_hooks)
+        return user_hooks
+
+    def add_api_hook(self, cb, module='', api_name='', argc=0, call_conv=None,
+                     emu=None) -> common.ApiHook:
         """
         Add an API level hook (e.g. kernel32.CreateFile) here
         """
+        module = module.lower()
+        api_name = api_name.lower()
+
+        wildcard_module, wildcard_api = False, False
+        for wc in ['?', '*', '[', ']']:
+            if wc in module:
+                wildcard_module = True
+            if wc in api_name:
+                wildcard_api = True
+
         if not emu:
             emu = self
         hook = common.ApiHook(emu, self.emu_eng, cb, module, api_name, argc, call_conv)
-        hl = self.hooks.get(common.HOOK_API)
-        if not hl:
-            self.hooks.update({common.HOOK_API: [hook, ]})
+        _hooks: MODULE_LEVEL = self.hooks.get(common.HOOK_API)
+
+        api_dictionary = (
+            {
+              api_name: [hook]
+            },
+            wildcard_api)
+        if not _hooks:
+            # First addition
+            obj = ({module: api_dictionary}, wildcard_module)
         else:
-            hl.append(hook)
+            module_dict, previous_wildcard_module = _hooks
+            try:
+                api_dict, previous_wildcard_api = module_dict[module]
+            except KeyError:
+                # The module asked is not present, so we just add the api dictionary
+                module_dict[module] = api_dictionary
+            else:
+                # The module asked is present, so we can just add the hook
+                api_dict.setdefault(api_name, []).append(hook)
+                module_dict[module] = (api_dict, previous_wildcard_api | wildcard_api)
+            obj = (module_dict, previous_wildcard_module | wildcard_module)
+        self.hooks.update({common.HOOK_API: obj})
+        return hook
 
     def add_code_hook(self, cb, begin=1, end=0, ctx={}, emu=None):
         """
@@ -911,6 +1007,24 @@ class BinaryEmulator(MemoryManager):
 
         return hook
 
+    def add_mem_map_hook(self, cb, begin=1, end=0, emu=None):
+        """
+        Add a hook that will fire for memory maps
+        """
+        if not emu:
+            emu = self
+        hook = common.MapMemHook(emu, self.emu_eng, cb, begin, end)
+        hl = self.hooks.get(common.HOOK_MEM_MAP)
+        if not hl:
+            self.hooks.update({common.HOOK_MEM_MAP: [hook, ]})
+        else:
+            hl.insert(0, hook)
+
+        if self.emu_eng:
+            hook.add()
+
+        return hook
+
     def _hook_mem_invalid_dispatch(self, emu, access, address, size, value, ctx):
         """
         This handler will dispatch other invalid memory hooks
@@ -966,3 +1080,21 @@ class BinaryEmulator(MemoryManager):
             hook.add()
 
         return hook
+
+    def add_instruction_hook(self, cb, begin=1, end=0, ctx=[], emu=None, insn=None):
+        """
+        Add a hook that will fire for IN, SYSCALL, or SYSENTER instructions
+        """
+        if not emu:
+            emu = self
+        hook = common.InstructionHook(emu, self.emu_eng, cb, ctx=[], insn=insn)
+        hl = self.hooks.get(common.HOOK_INSN)
+        if not hl:
+            self.hooks.update({common.HOOK_INSN: [hook, ]})
+        else:
+            hl.insert(0, hook)
+
+        if self.emu_eng:
+            hook.add()
+
+        return hook
