@@ -52,6 +52,13 @@ class Speakeasy(object):
         self.mem_write_hooks = []
         self.mem_invalid_hooks = []
         self.interrupt_hooks = []
+        self.mem_map_hooks = []
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        del self
 
     def _init_config(self, config: dict) -> None:
         """
@@ -136,8 +143,11 @@ class Speakeasy(object):
             h = self.interrupt_hooks.pop(0)
             cb, ctx = h
             self.add_interrupt_hook(cb, ctx)
+        while self.mem_map_hooks:
+            h = self.mem_map_hooks.pop(0)
+            self.add_mem_map_hook(h)
 
-    def disasm(self, addr: int, size: int):
+    def disasm(self, addr: int, size: int, fast=True):
         """
         Get the disassembly from an address
 
@@ -147,7 +157,10 @@ class Speakeasy(object):
         return:
             A tuple of: (mnemonic, operands, and the full instruction)
         """
-        return self.emu.get_disasm(addr, size)
+        try:
+            return self.emu.get_disasm(addr, size, fast)
+        except Exception:
+            raise SpeakeasyError("Failed to disassemble at address: 0x%x" % (addr))
 
     def is_pe(self, data: bytes) -> bool:
         """
@@ -268,7 +281,7 @@ class Speakeasy(object):
             module: name of the module containing the target API
             api_name: Name of the API to hook. Wild cards (e.g. *) are supported.
             argc: force the emulator to account for this amount of arguments (for stack cleanup)
-            call_conv: force the emulator to use the suppolied calling convention for this hook
+            call_conv: force the emulator to use the supplied calling convention for this hook
         return:
             Hook object for newly registered hooks
         """
@@ -278,12 +291,32 @@ class Speakeasy(object):
         return self.emu.add_api_hook(cb, module=module, api_name=api_name, argc=argc,
                                      call_conv=call_conv, emu=self)
 
+    def resume(self, addr, count=-1):
+        """
+        Resume emulating at the specified address
+
+        args:
+            addr: Address to being emulation at
+            count: number of instructions
+        return:
+            None
+        """
+        self.emu.run_complete = False
+        self.emu.resume(addr, count=count)
+
     def stop(self) -> None:
         """
         Stops emulation
         """
         return self.emu.stop()
 
+    def shutdown(self) -> None:
+        """
+        Closes the emulation instance
+        """
+        # TODO
+        return
+
     def call(self, addr: int, params=[]) -> None:
         """
         Start emulating at the specified address
@@ -345,6 +378,38 @@ class Speakeasy(object):
             return
         return self.emu.add_mem_write_hook(cb, begin=begin, end=end, emu=self)
 
+    def add_IN_instruction_hook(self, cb: Callable, begin=1, end=0):
+        """
+        Set a callback to fire when an IN instruction executes
+
+        args:
+            cb: Callable python function to execute
+            begin: beginning of the address range to hook
+            end: end of the address range to hook
+        return:
+            Hook object for newly registered hooks
+        """
+        if not self.emu:
+            self.mem_write_hooks.append((cb, begin, end))
+            return
+        return self.emu.add_instruction_hook(cb, begin=begin, end=end, emu=self, insn=218)
+
+    def add_SYSCALL_instruction_hook(self, cb: Callable, begin=1, end=0):
+        """
+        Set a callback to fire when a SYSCALL / SYSENTER instruction executes
+
+        args:
+            cb: Callable python function to execute
+            begin: beginning of the address range to hook
+            end: end of the address range to hook
+        return:
+            Hook object for newly registered hooks
+        """
+        if not self.emu:
+            self.mem_write_hooks.append((cb, begin, end))
+            return
+        return self.emu.add_instruction_hook(cb, begin=begin, end=end, emu=self, insn=700)
+
     def add_mem_invalid_hook(self, cb: Callable):
         """
         Get a callback for when a memory access violation occurs
@@ -397,6 +462,24 @@ class Speakeasy(object):
         """
         return self.emu.get_address_map(addr)
 
+    def get_user_modules(self) -> list:
+        """
+        Get the address ranges of loaded user modules
+
+        return:
+            List of all currently loaded user modules
+        """
+        return self.emu.get_user_modules()
+
+    def get_sys_modules(self) -> list:
+        """
+        Get the address ranges of loaded system modules
+
+        return:
+            List of all currently loaded system modules
+        """
+        return self.emu.get_sys_modules()
+
     def mem_alloc(self, size, base=None, tag='speakeasy.None') -> int:
         """
         Allocate a block of memory in the emulation space
@@ -431,7 +514,10 @@ class Speakeasy(object):
         return:
             Python bytes object contained the data read
         """
-        return self.emu.mem_read(addr, size)
+        try:
+            return self.emu.mem_read(addr, size)
+        except Exception:
+            raise SpeakeasyError("Failed to read %d bytes at address: 0x%x" % (size, addr))
 
     def mem_write(self, addr: int, data: bytes) -> None:
         """
@@ -443,7 +529,10 @@ class Speakeasy(object):
         return:
             None
         """
-        return self.emu.mem_write(addr, data)
+        try:
+            return self.emu.mem_write(addr, data)
+        except Exception:
+            raise SpeakeasyError("Failed to write %d bytes at address: 0x%x" % (len(data), addr))
 
     def mem_cast(self, obj, addr: int):
         """
@@ -559,6 +648,176 @@ class Speakeasy(object):
                 continue
             yield (tag, base, size, is_free, proc, data)
 
+    def read_mem_string(self, address: int, width=1, max_chars=0) -> str:
+        """
+        Read a string from emulated memory
+
+        args:
+            address: address of the string to read
+            width: character width
+            max_chars: maximum characters to read, 0 reads until null terminator
+
+        return:
+            decoded string
+        """
+        return self.emu.read_mem_string(address, width, max_chars)
+
+    def get_symbols(self) -> dict:
+        """
+        Returns a dictionary of symbol information
+
+        return:
+            a dictionary of symbol information
+        """
+        return self.emu.symbols
+
+    def get_ret_address(self) -> int:
+        """
+        Returns the value stored at the top of the stack
+
+        return:
+            value stored at the top of the stack
+        """
+        return self.emu.get_ret_address()
+
+    def push_stack(self, val: int) -> None:
+        """
+        Put a value on the stack and adjust the stack pointer
+
+        args:
+            val: value to push to the stack
+        return:
+            None
+        """
+        self.emu.push_stack(val)
+
+    def pop_stack(self) -> int:
+        """
+        Get value from the stack and adjust the stack pointer
+
+        return:
+            value stored at the top of the stack
+        """
+        return self.emu.pop_stack()
+
+    def get_stack_ptr(self) -> int:
+        """
+        Get the current address of the stack pointer
+
+        return:
+            address of stack pointer
+        """
+        return self.emu.get_stack_ptr()
+
+    def set_stack_ptr(self, addr: int) -> None:
+        """
+        Set the current address of the stack pointer
+
+        args:
+            addr: address to set the stack pointer to
+        return:
+            None
+        """
+        self.emu.set_stack_ptr(addr)
+
+    def set_pc(self, addr: int) -> None:
+        """
+        Set the value of the current program counter
+
+        args:
+            addr: address to set the program counter to
+        return:
+            None
+        """
+        self.emu.set_pc(addr)
+
+    def reset_stack(self, base: int) -> tuple:
+        """
+        Reset stack to the supplied base address
+
+        args:
+            base: stack base address
+        return:
+            base, ptr
+        """
+        return self.emu.reset_stack(base)
+
+    def get_stack_base(self) -> int:
+        """
+        Get the base address of the stack
+
+        return:
+            base address of stack
+        """
+        return self.emu.stack_base
+
+    def get_arch(self) -> int:
+        """
+        Get the architecture of the emulator
+
+        return:
+            emulator architecture constant value
+        """
+        return self.emu.get_arch()
+
+    def get_ptr_size(self) -> int:
+        """
+        Get the size of a pointer
+
+        return:
+            pointer size
+        """
+        return self.emu.ptr_size
+
+    def get_all_registers(self) -> dict:
+        """
+        Get the state of all registers
+
+        return:
+            Dict containing emulation register states
+        """
+        return self.emu.get_register_state()
+
+    def get_symbol_from_address(self, address: int) -> str:
+        """
+        If the supplied address is related to a known symbol, look it up here
+
+        args:
+            address: address to lookup
+
+        return:
+            symbol name
+        """
+        return self.emu.get_symbol_from_address(address)
+
+    def is_address_valid(self, address: int) -> bool:
+        """
+        Was this address previously reserved or mapped?
+
+        args:
+            address: address to check
+
+        return:
+            True if address is valid, false otherwise
+        """
+        return self.emu.is_address_valid(address)
+
+    def add_mem_map_hook(self, cb: Callable, begin=1, end=0):
+        """
+        Set a callback to fire when a memory address is mapped
+
+        args:
+            cb: Callable python function to execute
+            begin: beginning of the address range to hook
+            end: end of the address range to hook
+        return:
+            Hook object for newly registered hooks
+        """
+        if not self.emu:
+            self.mem_map_hooks.append((cb, begin, end))
+            return
+        return self.emu.add_mem_map_hook(cb, begin=begin, end=end, emu=self)
+
     def create_memdump_archive(self) -> bytes:
         """
         Creates a memory dump archive package of the emulated sample.
