@@ -258,7 +258,7 @@ class Driver(KernelObject):
         ldte.write_back()
         return ldte
 
-    def init_driver_object(self, name=None, pe=None):
+    def init_driver_object(self, name=None, pe=None, is_decoy=True):
         """
         Initialize the DRIVER_OBJECT
         """
@@ -275,6 +275,15 @@ class Driver(KernelObject):
             drvobj.DriverSize = pe.image_size
             drvobj.DriverInit = pe.get_base() + pe.ep
 
+            if is_decoy:
+                ep = pe.get_base() + pe.ep
+                drvobj.MajorFunction[ddk.IRP_MJ_CREATE] = ep + 1
+                drvobj.MajorFunction[ddk.IRP_MJ_READ] = ep + 2
+                drvobj.MajorFunction[ddk.IRP_MJ_WRITE] = ep + 3
+                drvobj.MajorFunction[ddk.IRP_MJ_DEVICE_CONTROL] = ep + 4
+                drvobj.MajorFunction[ddk.IRP_MJ_PNP] = ep + 5
+                drvobj.MajorFunction[ddk.IRP_MJ_INTERNAL_DEVICE_CONTROL] = ep+6
+
             if not name:
                 drvname = os.path.splitext(pe.path)[0]
                 drvname = os.path.basename(drvname)
@@ -521,9 +530,8 @@ class Process(KernelObject):
         self.peb = PEB(emu=emu)
         self.peb_ldr_data = PebLdrData(self.emu)
         self.is_peb_active = False
-        self.set_process_parameters()
-
         self.path = path
+        self.set_process_parameters(emu)
         self.image = ''
         self.title = ''
 
@@ -538,10 +546,9 @@ class Process(KernelObject):
         self.peb.write_back()
         self.peb_ldr_data.address = addr
 
-    def set_process_parameters(self):
-        self.peb.object.ProcessParameters = self.emu.mem_map(0x40,
-                                                             tag=self.get_mem_tag() +
-                                                             '.ProcessParameters')
+    def set_process_parameters(self, emu):
+        process_parameters = RTL_USER_PROCESS_PARAMETERS(emu=emu, proc=self)
+        self.peb.object.ProcessParameters = process_parameters.address
         self.peb.write_back()
 
     def get_peb_ldr(self):
@@ -633,8 +640,11 @@ class Process(KernelObject):
         self.ldr_entries.append(ldte)
         first = self.ldr_entries[0]
 
-        ldte.object.DllBase = module.get_base()
+        ldte.object.InLoadOrderLinks.Flink = first.address
+        ldte.object.InMemoryOrderLinks.Flink = first.address + self.sizeof(list_type)
+        ldte.object.InInitializationOrderLinks.Flink = first.address + self.sizeof(list_type) * 2
 
+        ldte.object.DllBase = module.get_base()
         dllname = (module.get_emu_path() + '\x00').encode('utf-16le')
         name_addr = ldte.address + ldte.sizeof()
         self.emu.mem_write(name_addr, dllname)
@@ -658,24 +668,32 @@ class Process(KernelObject):
         prev.object.InMemoryOrderLinks.Flink = ldte.address + \
             self.sizeof(list_type)
 
-        imol = prev.object.InMemoryOrderLinks.Flink
-        prev.object.InInitializationOrderLinks.Flink = imol + \
-            self.sizeof(list_type)
+        if first is ldte:
+            prev.object.InInitializationOrderLinks.Flink = 0
+        else:
+            imol = prev.object.InMemoryOrderLinks.Flink
+            prev.object.InInitializationOrderLinks.Flink = imol + \
+                self.sizeof(list_type)
 
         ldte.object.InLoadOrderLinks.Blink = prev.address
         ldte.object.InMemoryOrderLinks.Blink = prev.address + \
             self.sizeof(list_type)
 
-        imol = ldte.object.InMemoryOrderLinks.Blink
-        ldte.object.InInitializationOrderLinks.Blink = imol + \
-            self.sizeof(list_type)
+        if first is ldte:
+            ldte.object.InInitializationOrderLinks.Blink = 0
+        else:
+            imol = ldte.object.InMemoryOrderLinks.Blink
+            ldte.object.InInitializationOrderLinks.Blink = imol + \
+                self.sizeof(list_type)
 
         prev.write_back()
         ldte.write_back()
 
         first.object.InLoadOrderLinks.Blink = prev.address
         first.object.InMemoryOrderLinks.Blink = prev.address
-        first.object.InInitializationOrderLinks.Blink = prev.address
+        if first is not ldte:
+            first.object.InInitializationOrderLinks.Blink = prev.address
+
         first.write_back()
 
         pld.object.InLoadOrderModuleList.Flink = first.address
@@ -683,9 +701,12 @@ class Process(KernelObject):
             pld.object.InLoadOrderModuleList.Flink + \
             self.sizeof(list_type)
 
-        pld.object.InInitializationOrderModuleList.Flink = \
-            pld.object.InMemoryOrderModuleList.Flink + \
-            self.sizeof(list_type)
+        # Lets just copy InMemoryOrderModuleList but skip the main EXE module
+        head = pld.object.InMemoryOrderModuleList.Flink
+        le = self.emu.mem_cast(ntoskrnl.LIST_ENTRY(self.emu.get_ptr_size()),
+                               head)
+
+        pld.object.InInitializationOrderModuleList.Flink = le.Flink + self.sizeof(list_type)
 
         pld.object.InLoadOrderModuleList.Blink = prev.address
         pld.object.InMemoryOrderModuleList.Blink = \
@@ -707,17 +728,45 @@ class Process(KernelObject):
             self.add_module_to_peb(mod)
 
 
+class RTL_USER_PROCESS_PARAMETERS(KernelObject):
+    def __init__(self, emu, proc):
+        super(RTL_USER_PROCESS_PARAMETERS, self).__init__(emu=emu)
+
+        self.object = self.nt_types.RTL_USER_PROCESS_PARAMETERS(emu.get_ptr_size())
+        proc_path = (proc.path + '\x00').encode('utf-16le')
+        proc_cmdline = (proc.cmdline + '\x00').encode('utf-16le')
+        size = self.sizeof()
+        size += len(proc_path)
+        size += len(proc_cmdline)
+        self.address = emu.mem_map(size,
+                                   tag=proc.get_mem_tag() + '.ProcessParameters')
+        emu.mem_write(self.address + self.sizeof(), proc_path)
+        emu.mem_write(self.address + self.sizeof() + len(proc_path), proc_cmdline)
+
+        self.object.ImagePathName.Length = len(proc_path) - 2
+        self.object.ImagePathName.MaxLength = len(proc_path)
+        self.object.ImagePathName.Buffer = self.address + self.sizeof()
+
+        self.object.CommandLine.Length = len(proc_cmdline) - 2
+        self.object.CommandLine.MaxLength = len(proc_cmdline)
+        self.object.CommandLine.Buffer = self.address + self.sizeof() + len(proc_path)
+        self.write_back()
+
+
 class PEB(KernelObject):
     """
     Represents the process environment block. This structure contains a large amount of
     fields that are used internally by Windows. Shellcode may parse this structure in
     order to resolve exported functions.
     """
-    def __init__(self, emu):
+    def __init__(self, emu, address=None):
         super(PEB, self).__init__(emu=emu)
 
         self.object = self.nt_types.PEB(emu.get_ptr_size())
-        self.address = emu.mem_map(self.sizeof(), tag=self.get_mem_tag())
+        if not address:
+            self.address = emu.mem_map(self.sizeof(), tag=self.get_mem_tag())
+        else:
+            self.address = address
 
 
 class TEB(KernelObject):
@@ -842,6 +891,28 @@ class ObjectManager(object):
         obj.ref_cnt += 1
         return obj
 
+    def remove_object(self, obj):
+        """
+        Remove an object from the object manager
+        """
+        addr = None
+        for a, o in self.objects.items():
+            if o == obj:
+                addr = a
+                break
+        if addr:
+            self.objects.pop(addr)
+
+    def dec_ref(self, obj):
+        """
+        Dereferece an object and remove it from the object manager when its reference count is 0
+        """
+        if hasattr(obj, 'ref_cnt'):
+            obj.ref_cnt -= 1
+            if obj.ref_cnt <= 0:
+                self.remove_object(obj)
+            return obj.ref_cnt
+
     def get_handle(self, obj):
         tmp = KernelObject.curr_handle
         KernelObject.curr_handle += 4
@@ -863,9 +934,12 @@ class ObjectManager(object):
 
     def get_object_from_name(self, name, check_symlinks=True):
 
+        if not name:
+            return None
         name = name.rstrip('\\')
         for a, o in self.objects.items():
-
+            if not o.name:
+                continue
             if o.name.lower() == name.lower():
                 return o
         if check_symlinks:
