@@ -17,6 +17,7 @@ import speakeasy.common as common
 from speakeasy.errors import Win32EmuError
 
 from speakeasy.windows.sessman import SessionManager
+from speakeasy.windows.com import COM
 
 
 DLL_PROCESS_DETACH = 0
@@ -38,6 +39,7 @@ class Win32Emulator(WindowsEmulator):
         self.heap_allocs = []
         self.argv = argv
         self.sessman = SessionManager(config)
+        self.com = COM(config)
 
     def get_argv(self):
         """
@@ -167,8 +169,8 @@ class Win32Emulator(WindowsEmulator):
 
         # Strings the initial buffer so that we can detect decoded strings later on
         if self.profiler and self.do_strings:
-            self.profiler.strings['ansi'] = self.get_ansi_strings(data)
-            self.profiler.strings['unicode'] = self.get_unicode_strings(data)
+            self.profiler.strings['ansi'] = [a[1] for a in self.get_ansi_strings(data)]
+            self.profiler.strings['unicode'] = [u[1] for u in self.get_unicode_strings(data)]
 
         # Set the emulated path
         emu_path = ''
@@ -213,11 +215,22 @@ class Win32Emulator(WindowsEmulator):
             self.stop()
             raise Win32EmuError('Module not found')
 
+        # Check if any TLS callbacks exist, these run before the module's entry point
+        tls = module.get_tls_callbacks()
+        for i, cb_addr in enumerate(tls):
+            base = module.get_base()
+            if base < cb_addr < base + module.get_image_size():
+                run = Run()
+                run.start_addr = cb_addr
+                run.type = 'tls_callback_%d' % (i)
+                run.args = [base, DLL_PROCESS_ATTACH, 0]
+                self.add_run(run)
+
+        # Queue up the module's main entry point
         ep = module.base + module.ep
 
         run = Run()
         run.start_addr = ep
-        run.instr_cnt = 0
 
         main_exe = None
         if not module.is_exe():
@@ -269,6 +282,10 @@ class Win32Emulator(WindowsEmulator):
             p = objman.Process(self, path=module.get_emu_path(), base=module.base,
                                pe=module, cmdline=self.command_line)
             self.curr_process = p
+            self.om.objects.update({p.address: p})
+            mm = self.get_address_map(module.base)
+            if mm:
+                mm.process = self.curr_process
 
         t = objman.Thread(self,
                           stack_base=self.stack_base,
@@ -297,9 +314,15 @@ class Win32Emulator(WindowsEmulator):
         """
         Load position independent code (i.e. shellcode) to prepare for emulation
         """
-        self.arch = arch
         sc_hash = None
 
+        if arch == 'x86':
+            arch = _arch.ARCH_X86
+        elif arch in ('x64', 'amd64'):
+            arch = _arch.ARCH_AMD64
+
+        self.arch = arch
+
         if data:
             sc_hash = hashlib.sha256()
             sc_hash.update(data)
@@ -347,8 +370,9 @@ class Win32Emulator(WindowsEmulator):
             self.profiler.add_input_metadata(self.input)
             # Strings the initial buffer so that we can detect decoded strings later on
             if self.do_strings:
-                self.profiler.strings['ansi'] = self.get_ansi_strings(sc)
-                self.profiler.strings['unicode'] = self.get_unicode_strings(sc)
+                self.profiler.strings['ansi'] = [a[1] for a in self.get_ansi_strings(sc)]
+                self.profiler.strings['unicode'] = [u[1] for u in self.get_unicode_strings(sc)]
+        self.setup()
 
         return sc_addr
 
@@ -366,8 +390,6 @@ class Win32Emulator(WindowsEmulator):
         if not target:
             raise Win32EmuError('Invalid shellcode address')
 
-        self.setup()
-
         stack_commit = 0x4000
 
         self.stack_base, stack_addr = self.alloc_stack(stack_commit)
@@ -429,9 +451,18 @@ class Win32Emulator(WindowsEmulator):
         peb = proc.get_peb()
         proc.is_peb_active = True
         peb.object.ImageBaseAddress = proc.base
+        peb.object.OSMajorVersion = self.osversion['major']
+        peb.object.OSMinorVersion = self.osversion['minor']
+        peb.object.OSBuildNumber = self.osversion['build']
         peb.write_back()
         return peb
 
+    def set_unhandled_exception_handler(self, handler_addr):
+        """
+        Establish a handler for unhandled exceptions that occur during emulation
+        """
+        self.unhandled_exception_filter = handler_addr
+
     def setup(self, stack_commit=0):
 
         # Set the emulator to run in protected mode
@@ -518,11 +549,14 @@ class Win32Emulator(WindowsEmulator):
             if proc_mod:
                 all_user_mods = [proc_mod] + self.config_user_modules
                 user_modules = self.init_user_modules(all_user_mods)
-
             else:
                 user_modules = self.init_user_modules(self.config_user_modules)
 
             self.user_modules += user_modules
+            # add sample to user modules list if it is a dll
+            if self.modules and not self.modules[0][0].is_exe():
+                self.user_modules.append(self.modules[0][0])
+
         return self.user_modules
 
     def exit_process(self):
@@ -572,8 +606,9 @@ class Win32Emulator(WindowsEmulator):
             self.emu_complete = True
             if self.do_strings and self.profiler:
                 dec_ansi, dec_unicode = self.get_mem_strings()
-                dec_ansi = [a for a in dec_ansi if a not in self.profiler.strings['ansi']]
-                dec_unicode = [u for u in dec_unicode if u not in self.profiler.strings['unicode']]
+                dec_ansi = [a[1] for a in dec_ansi if a not in self.profiler.strings['ansi']]
+                dec_unicode = [u[1] for u in dec_unicode
+                               if u not in self.profiler.strings['unicode']]
                 self.profiler.decoded_strings['ansi'] = dec_ansi
                 self.profiler.decoded_strings['unicode'] = dec_unicode
         self.stop()
