@@ -16,6 +16,8 @@ from speakeasy.windows.regman import RegistryManager
 from speakeasy.windows.fileman import FileManager
 from speakeasy.windows.cryptman import CryptoManager
 from speakeasy.windows.netman import NetworkManager
+from speakeasy.windows.hammer import ApiHammer
+from speakeasy.windows.driveman import DriveManager
 
 import speakeasy.winenv.defs.nt.ddk as ddk
 import speakeasy.winenv.defs.windows.windows as windef
@@ -52,7 +54,7 @@ class WindowsEmulator(BinaryEmulator):
         self.page_size = 4096
         self.ptr_size = None
         self.user_modules = []
-        self.max_runs = 10
+        self.max_runs = 100
 
         self.sys_modules = []
         self.symbols = {}
@@ -61,6 +63,7 @@ class WindowsEmulator(BinaryEmulator):
         self.tmp_maps = []
         self.impdata_queue = []
         self.run_queue = []
+        self.suspended_runs = []
         self.cd = ''
         self.emu_hooks_set = True
         self.api = None
@@ -83,6 +86,7 @@ class WindowsEmulator(BinaryEmulator):
         self.curr_thread = None
         self.curr_exception_code = 0
         self.prev_pc = 0
+        self.unhandled_exception_filter = 0
 
         self.fs_addr = 0
         self.gs_addr = 0
@@ -96,7 +100,9 @@ class WindowsEmulator(BinaryEmulator):
         self.regman = RegistryManager(self.get_registry_config())
         self.fileman = FileManager(self.get_filesystem_config())
         self.netman = NetworkManager(config=self.get_network_config())
+        self.driveman = DriveManager(config=self.get_drive_config())
         self.cryptman = CryptoManager()
+        self.hammer = ApiHammer(self)
 
     def _parse_config(self, config):
         """
@@ -128,6 +134,7 @@ class WindowsEmulator(BinaryEmulator):
         self.do_strings = self.config_analysis.get('strings', False)
         self.registry_config = self.config.get('registry', {})
         self.modules_always_exist = self.config_modules.get('modules_always_exist', False)
+        self.functions_always_exist = self.config_modules.get('functions_always_exist', False)
 
     def get_registry_config(self):
         """
@@ -145,6 +152,9 @@ class WindowsEmulator(BinaryEmulator):
         raise NotImplementedError()
 
     def enable_code_hook(self):
+        if not self.tmp_code_hook and not self.mem_tracing_enabled:
+            self.tmp_code_hook = self.add_code_hook(cb=self._hook_code)
+
         if self.tmp_code_hook:
             self.tmp_code_hook.enable()
 
@@ -219,6 +229,12 @@ class WindowsEmulator(BinaryEmulator):
         """
         return self.fileman.get_file_from_handle(handle)
 
+    def file_delete(self, path):
+        """
+        Delete a file
+        """
+        return self.fileman.delete_file(path)
+
     def pipe_get(self, handle):
         """
         Get a pipe object from a handle
@@ -243,6 +259,12 @@ class WindowsEmulator(BinaryEmulator):
         """
         return self.cryptman
 
+    def get_drive_manager(self):
+        """
+        Get the drive manager
+        """
+        return self.driveman
+
     def reg_open_key(self, path, create=False):
         """
         Open or create a registry key in the emulation space
@@ -263,6 +285,12 @@ class WindowsEmulator(BinaryEmulator):
             return self.regman.get_key_from_path(path)
         return self.regman.get_key_from_handle(handle)
 
+    def reg_create_key(self, path):
+        """
+        Create a registry key
+        """
+        return self.regman.create_key(path)
+
     def _set_emu_hooks(self):
         """
         Unmap reserved memory space so we can handle events (e.g. import APIs,
@@ -370,6 +398,13 @@ class WindowsEmulator(BinaryEmulator):
             self.mem_map(self.page_size, base=0xFFFFF78000000000,
                          tag='emu.struct.KUSER_SHARED_DATA')
 
+    def resume(self, addr, count=-1):
+        """
+        Resume emulation at the specified address.
+        """
+        self.emu_eng.start(addr, timeout=self.timeout,
+                           count=count)
+
     def start(self):
         """
         Begin emulation executing each run in the specified run queue
@@ -458,7 +493,10 @@ class WindowsEmulator(BinaryEmulator):
         """
         Terminate a process (i.e. remove it from the known process list)
         """
-        self.processes.remove(proc)
+        try:
+            self.processes.remove(proc)
+        except ValueError:
+            pass
 
     def get_current_thread(self):
         """
@@ -633,7 +671,7 @@ class WindowsEmulator(BinaryEmulator):
         """
         Map the specified PE into the emulation space
         """
-        image_size = len(pe.mapped_image)
+        image_size = pe.image_size
         base = pe.base
         ranges = self.get_valid_ranges(image_size, addr=base)
         base, size = ranges
@@ -719,6 +757,9 @@ class WindowsEmulator(BinaryEmulator):
     def get_env(self):
         return self.env
 
+    def set_env(self, var, val):
+        return self.env.update({var.lower(): val})
+
     def get_os_version(self):
         return self.osversion
 
@@ -799,7 +840,7 @@ class WindowsEmulator(BinaryEmulator):
         self.processes.append(p)
         return p
 
-    def create_thread(self, addr, ctx, proc_obj, thread_type='thread'):
+    def create_thread(self, addr, ctx, proc_obj, thread_type='thread', is_suspended=False):
         """
         Create a thread object that will exist in the emulator
         """
@@ -807,6 +848,7 @@ class WindowsEmulator(BinaryEmulator):
             return 0, None
 
         thread = self.om.new_object(objman.Thread)
+        thread.process = proc_obj
         hnd = self.om.get_handle(thread)
 
         run = Run()
@@ -817,11 +859,25 @@ class WindowsEmulator(BinaryEmulator):
         run.process_context = proc_obj
         run.thread = thread
 
-        self.run_queue.append(run)
+        if not is_suspended:
+            self.run_queue.append(run)
+        else:
+            self.suspended_runs.append(run)
 
         # Returns handle
         return hnd, thread
 
+    def resume_thread(self, thread):
+        """
+        Resume a previously suspended thread
+        """
+        for r in self.suspended_runs:
+            if r.thread == thread:
+                _run = self.suspended_runs.pop(self.suspended_runs.index(r))
+                self.run_queue.append(_run)
+                return True
+        return False
+
     def get_dyn_imports(self):
         return self.dyn_imps
 
@@ -919,8 +975,9 @@ class WindowsEmulator(BinaryEmulator):
                 self._unset_emu_hooks()
                 return True
 
+        # Are there any SEH handlers registered?
         if self.dispatch_handlers:
-            rv = self.dispatch_seh(ddk.STATUS_ACCESS_VIOLATION)
+            rv = self.dispatch_seh(ddk.STATUS_ACCESS_VIOLATION, address)
             if rv:
                 return True
 
@@ -937,9 +994,10 @@ class WindowsEmulator(BinaryEmulator):
         """
         Collect emulator state information in the event of an error
         """
+        run = self.get_current_run()
         pc = self.get_pc()
         error = {}
-        self.log_error('Caught error: %s' % (desc))
+        self.log_error('0x%x: %s: Caught error: %s' % (pc, run.type, desc))
         error['type'] = desc
         error['pc'] = hex(pc)
         error['address'] = hex(address)
@@ -979,12 +1037,10 @@ class WindowsEmulator(BinaryEmulator):
                 name = 'Zw' + name[2:]
                 alt_imp_api = 'Zw%s' % (name[2:])
 
-        # Funnel CRTs into a single handler
-        if (dll.lower().startswith(('api-ms-win-crt', 'vcruntime', 'ucrtbased'))):
-            alt_imp_dll = 'msvcrt'
+        alt_imp_dll = winemu.normalize_dll_name(dll)
 
         # Bridge ntdll funcs to ntoskrnl if supported
-        elif dll.lower().startswith('ntdll'):
+        if dll.lower().startswith('ntdll'):
             alt_imp_dll = 'ntoskrnl'
             mod, func_attrs = self.api.get_export_func_handler(alt_imp_dll,
                                                                name)
@@ -1043,6 +1099,7 @@ class WindowsEmulator(BinaryEmulator):
         Forward imported functions to the corresponding handler (if any).
         """
         imp_api = '%s.%s' % (dll, name)
+        oret = self.get_ret_address()
         mod, func_attrs = self.api.get_export_func_handler(dll, name)
         if not func_attrs:
             mod, func_attrs = self.normalize_import_miss(dll, name)
@@ -1057,17 +1114,22 @@ class WindowsEmulator(BinaryEmulator):
             imp_api = '%s.%s' % (dll, name)
             default_ctx = {'func_name': imp_api}
 
-            hook = self.get_api_hook(dll, name)
-            if hook:
+            self.hammer.handle_import_func(imp_api, conv, argc)
+            hooks = self.get_api_hooks(dll, name)
+            if hooks:
                 from types import MethodType
                 hooked_func = MethodType(func, mod)
                 orig = lambda args: hooked_func(self, args, default_ctx) # noqa
-                rv = hook.cb(self, imp_api, orig, argv)
+                for hook in hooks:
+                    # each hook is called with the arguments, and only the last return value is
+                    # considered
+                    rv = hook.cb(self, imp_api, orig, argv)
             else:
                 try:
                     rv = self.api.call_api_func(mod, func, argv, ctx=default_ctx)
                 except Exception as e:
-                    self.log_exception('Error while calling API handler for %s:' % (imp_api))
+                    self.log_exception('0x%x: Error while calling API handler for %s:' %
+                                       (oret, imp_api))
                     error = self.get_error_info(str(e), self.get_pc(),
                                                 traceback=traceback.format_exc())
                     self.curr_run.error = error
@@ -1078,34 +1140,47 @@ class WindowsEmulator(BinaryEmulator):
             mm = self.get_address_map(ret)
 
             # Is this function being called from a dynamcially allocated memory segment?
-            if 'virtualalloc' in mm.get_tag().lower():
+            if mm and 'virtualalloc' in mm.get_tag().lower():
                 self._dynamic_code_cb(self, ret, 0, {})
 
             # Log the API args and return value
-            self.log_api(ret, imp_api, rv, argv)
+            self.log_api(oret, imp_api, rv, argv)
 
-            if not self.run_complete:
-                self.do_call_return(ret, argc, rv, conv=conv)
+            if not self.run_complete and ret == oret:
+                self.do_call_return(argc, ret, rv, conv=conv)
 
         else:
             # See if a user defined a hook for this unsupported function
-            hook = self.get_api_hook(dll, name)
-            if hook:
+            hooks = self.get_api_hooks(dll, name)
+            if hooks:
+                # Since the function is unsupported, just call the most accurate defined hook
+                hook = hooks[0]
                 imp_api = '%s.%s' % (dll, name)
 
                 if hook.call_conv is None:
                     hook.call_conv = _arch.CALL_CONV_STDCALL
 
                 argv = self.get_func_argv(hook.call_conv, hook.argc)
+                self.hammer.handle_import_func(imp_api, hook.call_conv, hook.argc)
                 rv = hook.cb(self, imp_api, None, argv)
                 ret = self.get_ret_address()
                 self.log_api(ret, imp_api, rv, argv)
-                self.do_call_return(ret, hook.argc, rv, conv=hook.call_conv)
+                self.do_call_return(hook.argc, ret, rv, conv=hook.call_conv)
+                return
+            elif self.functions_always_exist:
+                imp_api = '%s.%s' % (dll, name)
+                conv = _arch.CALL_CONV_STDCALL
+                argc = 4
+                argv = self.get_func_argv(conv, argc)
+                rv = 1
+                ret = self.get_ret_address()
+                self.log_api(ret, imp_api, rv, argv)
+                self.do_call_return(argc, ret, rv, conv=conv)
                 return
 
             run = self.get_current_run()
             error = self.get_error_info('unsupported_api', self.get_pc())
-            self.log_error("Unsupported API: %s" % (imp_api))
+            self.log_error("Unsupported API: %s (ret: 0x%x)" % (imp_api, oret))
             error['api_name'] = imp_api
             self.curr_run.error = error
             self.on_run_complete()
@@ -1124,7 +1199,6 @@ class WindowsEmulator(BinaryEmulator):
         High level function used to catch all invalid memory accesses that occur during
         emulation
         """
-
         try:
             access = self.emu_eng.mem_access.get(access)
             self.prev_pc = self.get_pc()
@@ -1140,6 +1214,13 @@ class WindowsEmulator(BinaryEmulator):
                     self.continue_seh()
                     self._unset_emu_hooks()
                     return True
+                elif address == winemu.API_CALLBACK_HANDLER_ADDR:
+                    run = self.get_current_run()
+                    if run.api_callbacks:
+                        pc, orig_func, args = run.api_callbacks.pop(0)
+                        self.do_call_return(len(args), pc)
+                        self._unset_emu_hooks()
+                    return True
                 return self._handle_invalid_fetch(emu, address, size,
                                                   value, ctx)
 
@@ -1317,7 +1398,7 @@ class WindowsEmulator(BinaryEmulator):
             return True
 
         if self.dispatch_handlers:
-            rv = self.dispatch_seh(ddk.STATUS_ACCESS_VIOLATION)
+            rv = self.dispatch_seh(ddk.STATUS_ACCESS_VIOLATION, address)
             if rv:
                 return True
         fakeout = address & 0xFFFFFFFFFFFFF000
@@ -1356,8 +1437,13 @@ class WindowsEmulator(BinaryEmulator):
         """
         Called when non-writable address is written to
         """
+        # ignore patches to APIs
+        if address >= winemu.EMU_RESERVED and \
+                address <= (winemu.EMU_RESERVED + winemu.EMU_RESERVE_SIZE):
+            return True
+
         if self.dispatch_handlers:
-            rv = self.dispatch_seh(ddk.STATUS_ACCESS_VIOLATION)
+            rv = self.dispatch_seh(ddk.STATUS_ACCESS_VIOLATION, address)
             if rv:
                 return True
 
@@ -1377,6 +1463,9 @@ class WindowsEmulator(BinaryEmulator):
         stay out of this hook as much as possible for speed considerations.
         """
 
+        if self.debug:
+            x = self.get_disasm(addr, size)[2]
+            print('0x%x: %s, edi=0x%x : esi=0x%x : ebp=0x%x : eax=0x%x' % (addr, x, self.reg_read('edi'), self.reg_read('esi'), self.reg_read('ebp'), self.reg_read('eax'))) # noqa
         try:
             if self.curr_exception_code != 0:
                 self.dispatch_seh(self.curr_exception_code)
@@ -1415,12 +1504,10 @@ class WindowsEmulator(BinaryEmulator):
                                                  'little'))
                 return True
 
-            # x = self.get_disasm(addr, size)[2]
-            # print('0x%x: %s'% (addr, x))
-
             if not self.mem_tracing_enabled:
                 # Disabling the code hook here grants a significant speed bump
-                self.disable_code_hook()
+                if not self.debug:
+                    self.disable_code_hook()
 
             if self.max_instructions != -1 and self.curr_run.instr_cnt >= self.max_instructions:
                 self.on_run_complete()
@@ -1511,12 +1598,12 @@ class WindowsEmulator(BinaryEmulator):
         Load a new module into the emulation space if necessary
         """
         ums = self.get_user_modules()
-        lib = os.path.splitext(mod_name)[0]
+        lib = ntpath.basename(mod_name)
+        lib = os.path.splitext(lib)[0]
 
         for um in ums:
             base = ntpath.basename(um.get_emu_path())
             base = os.path.splitext(base)[0]
-
             if lib.lower() == base.lower():
                 hmod = um.get_base()
                 return hmod
@@ -1532,7 +1619,8 @@ class WindowsEmulator(BinaryEmulator):
 
         # Add the newly loaded module to the current process's PEB module list
         proc = self.get_current_process()
-        proc.add_module_to_peb(mod)
+        if self.get_address_map(proc.get_peb_ldr().address):
+            proc.add_module_to_peb(mod)
 
         return mod.get_base()
 
@@ -1551,7 +1639,7 @@ class WindowsEmulator(BinaryEmulator):
             jit = winemu.JitPeFile(self.get_arch())
 
             funcs = [(f[4], f[0]) for k, f in mod_handler.funcs.items() if isinstance(k, str)]
-
+            data_exports = [k for k, d in mod_handler.data.items() if isinstance(k, str)]
             new = funcs.copy()
 
             if modname == 'ntdll':
@@ -1573,6 +1661,7 @@ class WindowsEmulator(BinaryEmulator):
             func_names = new
 
             func_names = [fn for o, fn in func_names]
+            func_names.sort()
 
             exports = []
             ords = [o for o, fn in funcs if o is not None]
@@ -1593,8 +1682,9 @@ class WindowsEmulator(BinaryEmulator):
             if not exports:
                 exports = func_names
 
+            exports += data_exports
             img = jit.get_decoy_pe_image(modname, exports)
-            mod = winemu.DecoyModule(data=img)
+            mod = winemu.DecoyModule(data=img, is_jitted=True)
 
             return mod
         return None
@@ -1618,7 +1708,7 @@ class WindowsEmulator(BinaryEmulator):
         for img in images:
             arch = img['arch']
             if arch == self.get_arch():
-                path = img['path']
+                path = self.get_native_module_path(mod_name=img['name'])
 
         if not path:
             path = default_file_path
@@ -1629,7 +1719,7 @@ class WindowsEmulator(BinaryEmulator):
         if isinstance(base, str):
             base = int(base, 16)
 
-        mod.decoy_path = modconf.get('path', emu_path)
+        mod.decoy_path = modconf.get('path', emu_path) or (name + '.dll')
         # Reserve memory for the module
         res, size = self.get_valid_ranges(mod.image_size,
                                           base)
@@ -1638,6 +1728,9 @@ class WindowsEmulator(BinaryEmulator):
         self.mem_reserve(size, base=res, tag='emu.module.%s' % (mod.name),
                          perms=common.PERM_MEM_RW)
 
+        if mod.decoy_path == '' and name != '':
+            mod.decoy_path = self.config.get('current_dir', 'C:\\Windows\\system32') + '\\' + name
+
         mod.base_name = ntpath.basename(mod.decoy_path)
 
         return mod
@@ -1671,7 +1764,6 @@ class WindowsEmulator(BinaryEmulator):
         parse a PE's export table while resolving exported functions
         """
         if not decoy.is_mapped:
-
             decoy.full_load()
 
             for exp in decoy.get_exports():
@@ -1840,10 +1932,18 @@ class WindowsEmulator(BinaryEmulator):
                                            "registers": regs,
                                            })
 
+            # EBX clobber, -1 is what I observed inside a VM
+            self.reg_write(_arch.X86_REG_EBX, 0xffffffff)
             self.set_pc(entry.Handler)
             return True
         return False
 
+    def get_reserved_ranges(self):
+        """
+        Get the allocated memory ranges that the emulator reserves
+        """
+        return (winemu.EMU_RESERVED, winemu.EMU_RESERVED_END)
+
     def _continue_seh_x86(self):
         """
         Get the next exception handler while processing SEH
@@ -1856,6 +1956,15 @@ class WindowsEmulator(BinaryEmulator):
         if seh.handler_ret_val is None:
             seh.handler_ret_val = ret_val
 
+        ctx = seh.get_context()
+
+        if seh.context_address:
+            ctx = self.mem_cast(ctx, seh.context_address)
+
+        # Always restore thread context, is it correct to always
+        # do this?
+        self.load_thread_context(ctx)
+
         for frame in seh.get_frames():
             if not frame.searched:
                 seh.set_current_frame(frame)
@@ -1897,9 +2006,40 @@ class WindowsEmulator(BinaryEmulator):
 
         self.run_complete = True
 
-    def dispatch_seh(self, except_code):
+    def dispatch_seh(self, except_code, faulting_address=None):
+        rv = False
         if self.get_arch() == _arch.ARCH_X86:
-            return self._dispatch_seh_x86(except_code)
+            rv = self._dispatch_seh_x86(except_code)
+        if not rv and self.unhandled_exception_filter:
+            # Create the _EXCEPTION_RECORD
+            record = self.wintypes.EXCEPTION_RECORD(self.get_ptr_size())
+            record.ExceptionCode = except_code
+            record.ExceptionFlags = 0
+            record.ExceptionAddress = self.get_pc()
+            record.NumberParameters = 0
+
+            exp_ptrs = self.wintypes.EXCEPTION_POINTERS(self.get_ptr_size())
+            p_exp_ptrs = self.mem_map(exp_ptrs.sizeof(), tag='emu.struct.EXCEPTION_POINTERS')
+            prec = self.mem_map(record.sizeof(), tag='emu.struct.EXCEPTION_RECORD')
+            pctx = self.mem_map(record.sizeof(), tag='emu.struct.EXCEPTION_CONTEXT')
+
+            exp_ptrs.ExceptionRecord = prec
+            exp_ptrs.ContextRecord = pctx
+
+            self.mem_write(p_exp_ptrs, exp_ptrs.get_bytes())
+            self.mem_write(prec, record.get_bytes())
+
+            sp = self.get_stack_ptr()
+            args = [p_exp_ptrs]
+            self.set_func_args(sp, winemu.EMU_RETURN_ADDR, *args)
+            self.set_pc(self.unhandled_exception_filter)
+            self.unhandled_exception_filter = 0
+            if faulting_address:
+                fakeout = faulting_address & 0xFFFFFFFFFFFFF000
+                self.mem_map(self.page_size, base=fakeout)
+                self.tmp_maps.append((fakeout, self.page_size))
+            rv = True
+        return rv
 
     def continue_seh(self):
         if self.get_arch() == _arch.ARCH_X86:
@@ -1914,10 +2054,18 @@ class WindowsEmulator(BinaryEmulator):
         hnd = self.om.get_handle(evt)
         return hnd, evt
 
+    def dec_ref(self, obj):
+        """
+        Dereference an object
+        """
+        return self.om.dec_ref(obj)
+
     def create_mutant(self, name=''):
         """
         Create a kernel mutant object
         """
+        if name == 0:
+            name = ''
         mtx = self.new_object(objman.Mutant)
         mtx.name = name
         hnd = self.om.get_handle(mtx)
@@ -1936,7 +2084,7 @@ class WindowsEmulator(BinaryEmulator):
         exception_list = self._get_exception_list()
         if exception_list and self.dispatch_handlers:
             # Catch software breakpoint interrupts
-            if intnum == 3:
+            if intnum == 3 or intnum == 0x2d:
                 self.curr_exception_code = ddk.STATUS_BREAKPOINT
                 self.prev_pc = self.get_pc()
                 self.enable_code_hook()
