@@ -8,6 +8,7 @@ import lznt1
 import speakeasy.winenv.arch as _arch
 import speakeasy.winenv.defs.nt.ddk as ddk
 import speakeasy.winenv.defs.registry.reg as regdefs
+import speakeasy.winenv.defs.windows.windows as windefs
 import speakeasy.winenv.defs.nt.ntoskrnl as ntos
 from speakeasy.errors import ApiEmuError
 from speakeasy.winenv.api import api
@@ -66,21 +67,13 @@ class Ntoskrnl(api.ApiHandler):
     @impdata('KeServiceDescriptorTable')
     def KeServiceDescriptorTable(self, ptr=0):
         """Kernel table containing SSDT"""
-        ssdt = self.win.SSDT(self.emu.get_ptr_size())
-        size = self.sizeof(ssdt)
-        size += (self.get_ptr_size() * 256)
-        desc_tbl = ptr
+        return self.emu.get_ssdt_ptr()
 
+    @impdata('KdDebuggerEnabled')
+    def KdDebuggerEnabled(self, ptr=0):
         if not ptr:
-            desc_tbl = self.mem_alloc(size, base=None,
-                                      tag='emu.struct.SSDT')
-
-        ssdt.NumberOfServices = 256
-        ssdt.pServiceTable = desc_tbl + self.sizeof(ssdt)
-
-        self.mem_write(desc_tbl, self.get_bytes(ssdt))
-
-        return desc_tbl
+            return self.mem_alloc(8, base=None, tag='emu.struct.KdDebuggerEnabled')
+        return ptr
 
     @apihook('ObfDereferenceObject', argc=1, conv=_arch.CALL_CONV_FASTCALL)
     def ObfDereferenceObject(self, emu, argv, ctx={}):
@@ -180,6 +173,10 @@ class Ntoskrnl(api.ApiHandler):
 
         return rv
 
+    @apihook('vsprintf_s', argc=4, conv=_arch.CALL_CONV_CDECL)
+    def vsprintf_s(self, emu, argv, ctx={}):
+        return self._vsnprintf(emu, argv, ctx)
+
     @apihook('RtlAnsiStringToUnicodeString', argc=3)
     def RtlAnsiStringToUnicodeString(self, emu, argv, ctx={}):
         """
@@ -363,6 +360,15 @@ class Ntoskrnl(api.ApiHandler):
         self.mem_write(dest, data)
         return dest
 
+    @apihook('IoDeleteDriver', argc=1)
+    def IoDeleteDriver(self, emu, argv, ctx={}):
+        """
+        VOID IoDeleteDriver(PDRIVER_OBJECT DriverObject)
+        """
+        drv, = argv
+
+        return
+
     @apihook('IoCreateDevice', argc=7)
     def IoCreateDevice(self, emu, argv, ctx={}):
         """
@@ -384,6 +390,8 @@ class Ntoskrnl(api.ApiHandler):
             name = self.read_unicode_string(name).replace('\x00', '')
 
         driver_obj = self.get_object_from_addr(drv)
+        if not driver_obj:
+            return ddk.STATUS_INVALID_PARAMETER
 
         dev = emu.create_device_object(name, driver_obj, ext_size, devtype,
                                        chars, tag='api.object')
@@ -565,9 +573,10 @@ class Ntoskrnl(api.ApiHandler):
 
         elif sysclass == ddk.SYSTEM_INFORMATION_CLASS.SystemKernelDebuggerInformation:
             if sysinfo and syslen >= 2:
-                out = (1).to_bytes(2, 'little')
+                out = b'\x00\x01'
+                size = len(out)
                 self.mem_write(sysinfo, out)
-                nts = ddk.STATUS_PORT_NOT_SET
+                nts = ddk.STATUS_SUCCESS
 
         elif sysclass == ddk.SYSTEM_INFORMATION_CLASS.SystemProcessInformation:
             procs = emu.get_processes()
@@ -742,6 +751,31 @@ class Ntoskrnl(api.ApiHandler):
         argv.append(fin)
         return len(fin)
 
+    @apihook('_snprintf', argc=_arch.VAR_ARGS, conv=_arch.CALL_CONV_CDECL)
+    def _snprintf(self, emu, argv, ctx={}):
+        """
+        int _snprintf(
+            char *buffer,
+            size_t count,
+            const char *format [,
+            argument] ...
+            );
+        """
+        buf, cnt, fmt = emu.get_func_argv(_arch.CALL_CONV_CDECL, 3)
+        fmt_str = self.read_string(fmt)
+        fmt_cnt = self.get_va_arg_count(fmt_str)
+        if not fmt_cnt:
+            self.write_string(fmt_str[:cnt - 1], buf)
+            return len(fmt_str)
+
+        _argv = emu.get_func_argv(_arch.CALL_CONV_CDECL, 3 + fmt_cnt)[3:]
+        fin = self.do_str_format(fmt_str, _argv)
+
+        self.write_string(fin, buf)
+        argv.clear()
+        argv.append(fin)
+        return len(fin)
+
     @apihook('wcslen', argc=1, conv=_arch.CALL_CONV_CDECL)
     def wcslen(self, emu, argv, ctx={}):
         """
@@ -1209,6 +1243,12 @@ class Ntoskrnl(api.ApiHandler):
         ProcessId, Process = argv
         rv = ddk.STATUS_SUCCESS
 
+        if ProcessId == 4:
+            proc = emu.get_system_process()
+            proc.ref_cnt += 1
+            self.mem_write(Process, proc.address.to_bytes(self.get_ptr_size(), 'little')) # noqa
+            return rv
+
         proc = self.get_object_from_id(ProcessId)
         if not proc:
             rv = ddk.STATUS_INVALID_CID
@@ -1365,6 +1405,9 @@ class Ntoskrnl(api.ApiHandler):
         size = int.from_bytes(self.mem_read(RegionSize,
                                             self.get_ptr_size()), 'little')
 
+        base = self.mem_read(BaseAddress, emu.get_ptr_size())
+        base = int.from_bytes(base, 'little')
+        argv[1] = '0x%x->0x%x' % (BaseAddress, base)
         base = self.mem_alloc(size, tag='api.virtalloc.%s' % obj.image, process=obj)
 
         emu._set_dyn_code_hook(base, size)
@@ -1623,6 +1666,22 @@ class Ntoskrnl(api.ApiHandler):
         rv = ddk.STATUS_DEBUGGER_INACTIVE
         return rv
 
+    @apihook('KdChangeOption', argc=0)
+    def KdChangeOption(self, emu, argv, ctx={}):
+        """
+        NTSTATUS KdChangeOption(
+          KD_OPTION Option,
+          ULONG     InBufferBytes,
+          PVOID     InBuffer,
+          ULONG     OutBufferBytes,
+          PVOID     OutBuffer,
+          PULONG    OutBufferNeeded
+        );
+        """
+
+        rv = ddk.STATUS_DEBUGGER_INACTIVE
+        return rv
+
     @apihook('MmGetSystemRoutineAddress', argc=1)
     def MmGetSystemRoutineAddress(self, emu, argv, ctx={}):
         """
@@ -1676,6 +1735,23 @@ class Ntoskrnl(api.ApiHandler):
         int_sys_time = int.from_bytes(sys_time, 'little')
         self.mem_write(LocalTime, int_sys_time.to_bytes(8, 'little'))
 
+    @apihook('CmRegisterCallbackEx', argc=6)
+    def CmRegisterCallbackEx(self, emu, argv, ctx={}):
+        """
+        NTSTATUS CmRegisterCallbackEx(
+            PEX_CALLBACK_FUNCTION Function,
+            PCUNICODE_STRING      Altitude,
+            PVOID                 Driver,
+            PVOID                 Context,
+            PLARGE_INTEGER        Cookie,
+            PVOID                 Reserved
+        );
+        """
+        # TODO: Emulate the callback routine
+        rv = ddk.STATUS_SUCCESS
+
+        return rv
+
     @apihook('CmRegisterCallback', argc=3)
     def CmRegisterCallback(self, emu, argv, ctx={}):
         """
@@ -1860,6 +1936,17 @@ class Ntoskrnl(api.ApiHandler):
 
         return
 
+    @apihook('ExReleaseFastMutex', argc=1)
+    def ExReleaseFastMutex(self, emu, argv, ctx={}):
+        """
+        VOID ExReleaseFastMutex(
+            _Inout_ PFAST_MUTEX FastMutex
+            );
+        """
+        FastMutex, = argv
+
+        return
+
     @apihook('ObfReferenceObject', argc=1)
     def ObfReferenceObject(self, emu, argv, ctx={}):
         """
@@ -2017,6 +2104,22 @@ class Ntoskrnl(api.ApiHandler):
         """
         return None
 
+    @apihook('RtlGetCompressionWorkSpaceSize', argc=3)
+    def RtlGetCompressionWorkSpaceSize(self, emu, argv, ctx={}):
+        """
+        NT_RTL_COMPRESS_API NTSTATUS RtlGetCompressionWorkSpaceSize(
+            USHORT CompressionFormatAndEngine,
+            PULONG CompressBufferWorkSpaceSize,
+            PULONG CompressFragmentWorkSpaceSize
+        );
+        """
+        engine, buffer_workspace, frag_workspace = argv
+        if buffer_workspace:
+            self.mem_write(buffer_workspace, 0x1000.to_bytes(4, 'little'))
+        if frag_workspace:
+            self.mem_write(frag_workspace, 0x1000.to_bytes(4, 'little'))
+        return ddk.STATUS_SUCCESS
+
     @apihook('RtlDecompressBuffer', argc=6)
     def RtlDecompressBuffer(self, emu, argv, ctx={}):
         """
@@ -2202,6 +2305,32 @@ class Ntoskrnl(api.ApiHandler):
 
         return nts
 
+    @apihook('ObGetFilterVersion', argc=0)
+    def ObGetFilterVersion(self, emu, argv, ctx={}):
+        """
+        NTKERNELAPI
+        USHORT
+        ObGetFilterVersion (
+            VOID
+            );
+        """
+        return 256
+
+    @apihook('ObRegisterCallbacks', argc=2)
+    def ObRegisterCallbacks(self, emu, argv, ctx={}):
+        """
+        NTKERNELAPI
+        NTSTATUS
+        ObRegisterCallbacks (
+            _In_ POB_CALLBACK_REGISTRATION CallbackRegistration,
+            _Outptr_ PVOID *RegistrationHandle
+            );
+        """
+        CallbackRegistration, RegistrationHandle = argv
+        nts = ddk.STATUS_SUCCESS
+
+        return nts
+
     @apihook('ZwDeleteKey', argc=1)
     def ZwDeleteKey(self, emu, argv, ctx={}):
         """
@@ -2382,6 +2511,77 @@ class Ntoskrnl(api.ApiHandler):
 
         return 0
 
+    @apihook('PsSetCreateProcessNotifyRoutineEx', argc=2)
+    def PsSetCreateProcessNotifyRoutineEx(self, emu, argv, ctx={}):
+        """
+        NTKERNELAPI
+        NTSTATUS
+        PsSetCreateProcessNotifyRoutineEx (
+            _In_ PCREATE_PROCESS_NOTIFY_ROUTINE_EX NotifyRoutine,
+            _In_ BOOLEAN Remove
+            );
+        """
+        NotifyRoutine, Remove = argv
+        rv = ddk.STATUS_SUCCESS
+
+        return rv
+
+    @apihook('PsSetLoadImageNotifyRoutine', argc=1)
+    def PsSetLoadImageNotifyRoutine(self, emu, argv, ctx={}):
+        """
+        NTKERNELAPI
+        NTSTATUS
+        PsSetLoadImageNotifyRoutine(
+            _In_ PLOAD_IMAGE_NOTIFY_ROUTINE NotifyRoutine
+            );
+        """
+        NotifyRoutine = argv # noqa
+        rv = ddk.STATUS_SUCCESS
+
+        return rv
+
+    @apihook('PsRemoveLoadImageNotifyRoutine', argc=1)
+    def PsRemoveLoadImageNotifyRoutine(self, emu, argv, ctx={}):
+        """
+        NTKERNELAPI
+        NTSTATUS
+        PsRemoveLoadImageNotifyRoutine(
+            _In_ PLOAD_IMAGE_NOTIFY_ROUTINE NotifyRoutine
+            );
+        """
+        NotifyRoutine = argv # noqa
+        rv = ddk.STATUS_SUCCESS
+
+        return rv
+
+    @apihook('PsSetCreateThreadNotifyRoutine', argc=1)
+    def PsSetCreateThreadNotifyRoutine(self, emu, argv, ctx={}):
+        """
+        NTKERNELAPI
+        NTSTATUS
+        PsSetCreateThreadNotifyRoutine (
+            _In_ PCREATE_THREAD_NOTIFY_ROUTINE NotifyRoutine
+            );
+        """
+        NotifyRoutine = argv # noqa
+        rv = ddk.STATUS_SUCCESS
+
+        return rv
+
+    @apihook('PsRemoveCreateThreadNotifyRoutine', argc=1)
+    def PsRemoveCreateThreadNotifyRoutine(self, emu, argv, ctx={}):
+        """
+        NTKERNELAPI
+        NTSTATUS
+        PsRemoveCreateThreadNotifyRoutine (
+            _In_ PCREATE_THREAD_NOTIFY_ROUTINE NotifyRoutine
+            );
+        """
+        NotifyRoutine = argv # noqa
+        rv = ddk.STATUS_SUCCESS
+
+        return rv
+
     @apihook('mbstowcs', argc=3)
     def mbstowcs(self, emu, argv, ctx={}):
         """
@@ -2558,7 +2758,7 @@ class Ntoskrnl(api.ApiHandler):
                 elif create_disp == ddk.FILE_OPEN:
                     nts = ddk.STATUS_UNSUCCESSFUL
                 elif create_disp == ddk.FILE_OPEN_IF:
-                    hfile = self.file_open(npath, create=False)
+                    hfile = self.file_open(npath, create=True)
                 elif create_disp == ddk.FILE_OVERWRITE:
                     nts = ddk.STATUS_UNSUCCESSFUL
                 elif create_disp == ddk.FILE_OVERWRITE_IF:
@@ -2711,6 +2911,7 @@ class Ntoskrnl(api.ApiHandler):
         );
         """
         FileHandle, evt, apc, apc_ctx, ios, buf, length, offset, key = argv
+        length = length & 0xFFFFFFFF
 
         nts = ddk.STATUS_INVALID_PARAMETER
         _file = self.file_get(FileHandle)
@@ -2820,6 +3021,7 @@ class Ntoskrnl(api.ApiHandler):
             size = int.from_bytes(size, 'little')
         hmap = fm.file_create_mapping(FileHandle, name, size, SectionPageProtection)
         self.mem_write(SectionHandle, hmap.to_bytes(self.get_ptr_size(), byteorder='little'))
+        argv[0] = hmap
 
         return ddk.STATUS_SUCCESS
 
@@ -2897,8 +3099,10 @@ class Ntoskrnl(api.ApiHandler):
                 fname = fname.replace('.', '_')
                 mm.update_tag('%s.%s.0x%x' % (tag_prefix, fname, buf))
                 self.mem_write(buf, data)
+                if ViewSize:
+                    self.mem_write(ViewSize, size.to_bytes(self.get_ptr_size(), 'little'))
                 argv[2] = buf
-                argv[3] = size
+                argv[6] = size
             elif not pref_address:
                 if bytes_to_map == 0:
                     bytes_to_map = sect.size
@@ -2907,8 +3111,10 @@ class Ntoskrnl(api.ApiHandler):
                                      process=proc_obj)
                 mm = emu.get_address_map(buf)
                 mm.update_tag('%s.0x%x' % (tag_prefix, buf))
+                if ViewSize:
+                    self.mem_write(ViewSize, size.to_bytes(self.get_ptr_size(), 'little'))
                 argv[2] = buf
-                argv[3] = size
+                argv[6] = size
                 sect.add_view(buf, full_offset, size, access)
             else:
                 buf = pref_address
@@ -2939,6 +3145,47 @@ class Ntoskrnl(api.ApiHandler):
 
         return block
 
+    @apihook('ZwGetContextThread', argc=2)
+    def ZwGetContextThread(self, emu, argv, ctx={}):
+        '''
+        BOOL ZwGetContextThread(
+            HANDLE    hThread,
+            LPCONTEXT lpContext
+        );
+        '''
+        hThread, lpContext = argv
+
+        obj = self.get_object_from_handle(hThread)
+        if not obj:
+            return False
+
+        context = obj.get_context()
+
+        self.mem_write(lpContext, context.get_bytes())
+
+        return True
+
+    @apihook('ZwSetContextThread', argc=2)
+    def ZwSetContextThread(self, emu, argv, ctx={}):
+        '''
+        BOOL ZwSetContextThread(
+            HANDLE    hThread,
+            LPCONTEXT lpContext
+        );
+        '''
+        hThread, lpContext = argv
+
+        obj = self.get_object_from_handle(hThread)
+        if not obj:
+            return False
+
+        context = windefs.CONTEXT(emu.get_ptr_size())
+        if lpContext:
+            _context = self.mem_cast(context, lpContext)
+            obj.set_context(_context)
+
+        return True
+
     @apihook('RtlFreeHeap', argc=3)
     def RtlFreeHeap(self, emu, argv, ctx={}):
         '''
