@@ -108,6 +108,22 @@ class Msvcrt(api.ApiHandler):
 
         return rv
 
+    @apihook('__wgetmainargs', argc=5)
+    def __wgetmainargs(self, emu, argv, ctx={}):
+        """
+        int __wgetmainargs (
+           int *_Argc,
+           wchar_t ***_Argv,
+           wchar_t ***_Env,
+           int _DoWildCard,
+           _startupinfo * _StartInfo);
+        """
+
+        _Argc, _Argv, _Env, _DoWildCard, _StartInfo = argv
+        rv = 0
+
+        return rv
+
     @apihook('__p___wargv', argc=0, conv=e_arch.CALL_CONV_CDECL)
     def __p___wargv(self, emu, argv, ctx={}):
         """WCHAR *** __p___wargv ()"""
@@ -356,6 +372,32 @@ class Msvcrt(api.ApiHandler):
 
         return ret
 
+    @apihook('wcsstr', argc=2, conv=e_arch.CALL_CONV_CDECL)
+    def wcsstr(self, emu, argv, ctx={}):
+        """
+        wchar_t *wcsstr(
+            const wchar_t *str,
+            const wchar_t *strSearch
+        );
+        """
+        hay, needle = argv
+
+        if hay:
+            _hay = self.read_mem_string(hay, 2)
+            argv[0] = _hay
+
+        if needle:
+            needle = self.read_mem_string(needle, 2)
+            argv[1] = needle
+
+        ret = _hay.find(needle)
+        if ret != -1:
+            ret = hay + ret
+        else:
+            ret = 0
+
+        return ret
+
     @apihook('strncat_s', argc=4, conv=e_arch.CALL_CONV_CDECL)
     def strncat_s(self, emu, argv, ctx={}):
         """
@@ -492,6 +534,31 @@ class Msvcrt(api.ApiHandler):
         argv.append(fin)
         return len(fin)
 
+    @apihook('_snprintf', argc=e_arch.VAR_ARGS, conv=e_arch.CALL_CONV_CDECL)
+    def _snprintf(self, emu, argv, ctx={}):
+        """
+        int _snprintf(
+        char *buffer,
+        size_t count,
+        const char *format [,
+        argument] ...
+        );
+        """
+        buf, count, fmt = emu.get_func_argv(e_arch.CALL_CONV_CDECL, 3)
+        fmt_str = self.read_string(fmt)
+        fmt_cnt = self.get_va_arg_count(fmt_str)
+        if not fmt_cnt:
+            self.write_string(fmt_str, buf)
+            return len(fmt_str)
+
+        _argv = emu.get_func_argv(e_arch.CALL_CONV_CDECL, 3 + fmt_cnt)[3:]
+        fin = self.do_str_format(fmt_str, _argv)
+
+        self.write_string(fin, buf)
+        argv.clear()
+        argv.append(fin)
+        return len(fin)
+
     @apihook('atoi', argc=1, conv=e_arch.CALL_CONV_CDECL)
     def atoi(self, emu, argv, ctx={}):
         """
@@ -531,6 +598,10 @@ class Msvcrt(api.ApiHandler):
         """
         return
 
+    @apihook('_set_app_type', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def _set_app_type(self, emu, argv, ctx={}):
+        return
+
     @apihook('__p__fmode', argc=0, conv=e_arch.CALL_CONV_CDECL)
     def __p__fmode(self, emu, argv, ctx={}):
         """
@@ -637,6 +708,29 @@ class Msvcrt(api.ApiHandler):
         self.mem_write(dest, data)
         return dest
 
+    @apihook('memcmp', argc=3, conv=e_arch.CALL_CONV_CDECL)
+    def memcmp(self, emu, argv, ctx={}):
+        """
+        int memcmp(
+           const void *buffer1,
+           const void *buffer2,
+           size_t count
+        );
+        """
+        diff = 0
+        buff1, buff2, cnt = argv
+        for i in range(cnt):
+            b1 = self.mem_read(buff1, 1)
+            b2 = self.mem_read(buff2, 1)
+            if b1 > b2:
+                diff = 1
+                break
+            elif b1 < b2:
+                diff = -1
+                break
+
+        return diff
+
     @apihook('_except_handler4_common', argc=6, conv=e_arch.CALL_CONV_CDECL)
     def _except_handler4_common(self, emu, argv, ctx={}):
         """
@@ -684,8 +778,14 @@ class Msvcrt(api.ApiHandler):
             rec = self.wintypes.EH4_SCOPETABLE_RECORD(emu.get_ptr_size())
             # The trylevel will tell us what scope record to get
             scope_record_offset = scope_table + st.sizeof()
-            scope_record_offset += (rec.sizeof() * reg.TryLevel)
+            tl = reg.TryLevel
+            if reg.TryLevel & 0x80000000:
+                tl = -0x100000000 + reg.TryLevel
 
+            if tl == -2:  # -2 is the outermost scope
+                tl = 0
+
+            scope_record_offset += (rec.sizeof() * tl)
             rec = self.mem_cast(rec, scope_record_offset)
 
             seh.add_frame(reg, st, [rec, ])
@@ -707,6 +807,19 @@ class Msvcrt(api.ApiHandler):
 
         return rv
 
+    @apihook('_except_handler3', argc=4, conv=e_arch.CALL_CONV_CDECL)
+    def _except_handler3(self, emu, argv, ctx={}):
+        """
+        int _except_handler3(
+        PEXCEPTION_RECORD exception_record,
+        PEXCEPTION_REGISTRATION registration,
+        PCONTEXT context,
+        PEXCEPTION_REGISTRATION dispatcher
+        );
+        """
+        rv = 1
+        return rv
+
     @apihook('_seh_filter_dll', argc=2, conv=e_arch.CALL_CONV_CDECL)
     def _seh_filter_dll(self, emu, argv, ctx={}):
         """
@@ -770,6 +883,33 @@ class Msvcrt(api.ApiHandler):
         chunk = self.heap_alloc(size, heap='HeapAlloc')
         return chunk
 
+    @apihook('calloc', argc=2, conv=e_arch.CALL_CONV_CDECL)
+    def calloc(self, emu, argv, ctx={}):
+        """
+        void *calloc(
+        size_t num,
+        size_t size
+        );
+        """
+        num, size, = argv
+
+        chunk = self.heap_alloc(num*size, heap='HeapAlloc')
+
+        buf = b'\x00' * (num*size)
+        self.mem_write(chunk, buf)
+
+        return chunk
+
+    @apihook('free', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def free(self, emu, argv, ctx={}):
+        """
+        void free(
+        void *memblock
+        );
+        """
+        mem, = argv
+        self.mem_free(mem)
+
     @apihook('_beginthreadex', argc=6, conv=e_arch.CALL_CONV_CDECL)
     def _beginthreadex(self, emu, argv, ctx={}):
         """
@@ -791,6 +931,20 @@ class Msvcrt(api.ApiHandler):
 
         return handle
 
+    @apihook('_beginthread', argc=3, conv=e_arch.CALL_CONV_CDECL)
+    def _beginthread(self, emu, argv, ctx={}):
+        """
+        uintptr_t _beginthread
+        void( __cdecl *start_address )( void * ),
+        unsigned stack_size,
+        void *arglist
+        );
+        """
+        start_address, stack_size, arglist = argv
+
+        handle, obj = self.create_thread(start_address, arglist, emu.get_current_process())
+        return handle
+
     @apihook('system', argc=1, conv=e_arch.CALL_CONV_CDECL)
     def system(self, emu, argv, ctx={}):
         """
@@ -869,7 +1023,7 @@ class Msvcrt(api.ApiHandler):
         new = (s1 + s2).encode('utf-16le')
         self.mem_write(_str1, new + b'\x00\x00')
         return _str1
-    
+
     @apihook('wcslen', argc=1, conv=e_arch.CALL_CONV_CDECL)
     def wcslen(self, emu, argv, ctx={}):
         """
@@ -950,6 +1104,104 @@ class Msvcrt(api.ApiHandler):
 
         return rv
 
+    @apihook('strcmp', argc=2, conv=e_arch.CALL_CONV_CDECL)
+    def strcmp(self, emu, argv, ctx={}):
+        """
+        int strcmp(
+            const char *string1,
+            const char *string2,
+        );
+        """
+        s1, s2 = argv
+        rv = 1
+
+        string1 = self.read_mem_string(s1, 1)
+        string2 = self.read_mem_string(s2, 1)
+        if string1 == string2:
+            rv = 0
+        argv[0] = string1
+        argv[1] = string2
+
+        return rv
+
+    @apihook('strrchr', argc=2, conv=e_arch.CALL_CONV_CDECL)
+    def strrchr(self, emu, argv, ctx={}):
+        """
+        char *strrchr(
+            const char *str,
+            int c
+            );
+        """
+        cstr, c = argv
+        cs = self.read_string(cstr)
+        hay = cs.encode('utf-8')
+        needle = c.to_bytes(1, 'little')
+
+        offset = hay.rfind(needle)
+        if offset < 0:
+            rv = 0
+        else:
+            rv = cstr + offset
+
+        argv[0] = cs
+        argv[1] = needle.decode('utf-8')
+
+        return rv
+
+    @apihook('_ftol', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def _ftol(self, emu, argv, ctx={}):
+        """
+        int _ftol(int);
+        """
+        f, = argv
+        return int(f)
+
+    @apihook('_adjust_fdiv', argc=0, conv=e_arch.CALL_CONV_CDECL)
+    def _adjust_fdiv(self, emu, argv, ctx={}):
+        """
+        void _adjust_fdiv(void)
+        """
+        return
+
+    @apihook('tolower', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def tolower(self, emu, argv, ctx={}):
+        """
+        int tolower ( int c );
+        """
+        c, = argv
+        return c | 0x20
+
+    @apihook('sscanf', argc=e_arch.VAR_ARGS, conv=e_arch.CALL_CONV_CDECL)
+    def sscanf(self, emu, argv, ctx={}):
+        """
+        int sscanf ( const char * s, const char * format, ...);
+        """
+        return
+
+    @apihook('strchr', argc=2, conv=e_arch.CALL_CONV_CDECL)
+    def strchr(self, emu, argv, ctx={}):
+        """
+        char *strchr(
+            const char *str,
+            int c
+            );
+        """
+        cstr, c = argv
+        cs = self.read_string(cstr)
+        hay = cs.encode('utf-8')
+        needle = c.to_bytes(1, 'little')
+
+        offset = hay.find(needle)
+        if offset < 0:
+            rv = 0
+        else:
+            rv = cstr + offset
+
+        argv[0] = cs
+        argv[1] = needle.decode('utf-8')
+
+        return rv
+
     @apihook('_set_invalid_parameter_handler', argc=1, conv=e_arch.CALL_CONV_CDECL)
     def _set_invalid_parameter_handler(self, emu, argv, ctx={}):
         """
@@ -961,6 +1213,19 @@ class Msvcrt(api.ApiHandler):
 
         return 0
 
+    @apihook('__CxxFrameHandler', argc=4, conv=e_arch.CALL_CONV_CDECL)
+    def __CxxFrameHandler(self, emu, argv, ctx={}):
+        """
+        EXCEPTION_DISPOSITION __CxxFrameHandler(
+            EHExceptionRecord  *pExcept,
+            EHRegistrationNode *pRN,
+            void               *pContext,
+            DispatcherContext  *pDC
+        )
+        """
+        pExcept, pRN, pContext, pDC, = argv
+        return 0
+
     @apihook('_vsnprintf', argc=4, conv=e_arch.CALL_CONV_CDECL)
     def _vsnprintf(self, emu, argv, ctx={}):
         """
@@ -988,3 +1253,163 @@ class Msvcrt(api.ApiHandler):
         argv[1] = fmt_str
 
         return rv
+
+    @apihook('__stdio_common_vsprintf', argc=7, conv=e_arch.CALL_CONV_CDECL)
+    def __stdio_common_vsprintf(self, emu, argv, ctx={}):
+        """
+        int __stdio_common_vsprintf(
+            unsigned int64 Options,
+            char *Buffer,
+            unsigned int BufferCount,
+            const char *format,
+            locale_t Locale,
+            va_list argptr
+        );
+        """
+        options_lo, options_hi, buffer, count, _format, locale, argptr = argv
+        rv = 0
+        fmt_str = self.read_mem_string(_format, 1)
+        fmt_cnt = self.get_va_arg_count(fmt_str)
+
+        vargs = self.va_args(argptr, fmt_cnt)
+
+        fin = self.do_str_format(fmt_str, vargs)
+        fin = fin[:count] + '\x00'
+
+        rv = len(fin)
+        self.mem_write(buffer, fin.encode('utf-8'))
+        argv[0] = fin.replace('\x00', '')
+        argv[1] = fmt_str
+
+        return rv
+
+    @apihook('_strcmpi', argc=2, conv=e_arch.CALL_CONV_CDECL)
+    def _strcmpi(self, emu, argv, ctx={}):
+        """
+        int _strcmpi(
+                const char *string1,
+                const char *string2
+                );
+        """
+        string1, string2 = argv
+        rv = 1
+
+        if not string1 or not string2:
+            return rv
+
+        cs1 = self.read_string(string1)
+        cs2 = self.read_string(string2)
+
+        argv[0] = cs1
+        argv[1] = cs2
+
+        if cs1.lower() == cs2.lower():
+            rv = 0
+
+        return rv
+
+    @apihook('_wcsicmp', argc=2, conv=e_arch.CALL_CONV_CDECL)
+    def _wcsicmp(self, emu, argv, ctx={}):
+        """
+        int _wcsicmp(
+                const wchar_t *string1,
+                const wchar_t *string2
+                );
+        """
+        string1, string2 = argv
+        rv = 1
+
+        if not string1 or not string2:
+            return rv
+
+        cs1 = self.read_wide_string(string1)
+        cs2 = self.read_wide_string(string2)
+
+        argv[0] = cs1
+        argv[1] = cs2
+
+        if cs1.lower() == cs2.lower():
+            rv = 0
+
+        return rv
+
+    @apihook('??3@YAXPAX@Z', argc=0, conv=e_arch.CALL_CONV_CDECL)
+    def __3_YAXPAX_Z(self, emu, argv, ctx={}):
+        return
+
+    @apihook('??2@YAPAXI@Z', argc=0, conv=e_arch.CALL_CONV_CDECL)
+    def __2_YAPAXI_Z(self, emu, argv, ctx={}):
+        return
+
+    @apihook('__current_exception_context', argc=0, conv=e_arch.CALL_CONV_CDECL)
+    def __current_exception_context(self, emu, argv, ctx={}):
+        return
+
+    @apihook('__current_exception', argc=0, conv=e_arch.CALL_CONV_CDECL)
+    def __current_exception(self, emu, argv, ctx={}):
+        return
+
+    @apihook('_set_new_mode', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def _set_new_mode(self, emu, argv, ctx={}):
+        return
+
+    @apihook('_configthreadlocale', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def _configthreadlocale(self, emu, argv, ctx={}):
+        return
+
+    @apihook('_setusermatherr', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def _setusermatherr(self, emu, argv, ctx={}):
+        return
+
+    @apihook('__setusermatherr', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def __setusermatherr(self, emu, argv, ctx={}):
+        return
+
+    @apihook('_cexit', argc=0, conv=e_arch.CALL_CONV_CDECL)
+    def _cexit(self, emu, argv, ctx={}):
+        # TODO: handle atexit flavor functions
+        self.exit_process()
+
+    @apihook('_c_exit', argc=0, conv=e_arch.CALL_CONV_CDECL)
+    def _c_exit(self, emu, argv, ctx={}):
+        self.exit_process()
+
+    @apihook('_register_thread_local_exe_atexit_callback', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def _register_thread_local_exe_atexit_callback(self, emu, argv, ctx={}):
+        return
+
+    @apihook('_crt_atexit', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def _crt_atexit(self, emu, argv, ctx={}):
+        return
+
+    @apihook('_controlfp_s', argc=3, conv=e_arch.CALL_CONV_CDECL)
+    def _controlfp_s(self, emu, argv, ctx={}):
+        return
+
+    @apihook('terminate', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def terminate(self, emu, argv, ctx={}):
+        self.exit_process()
+
+    @apihook('_crt_atexit', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def _crt_atexit(self, emu, argv, ctx={}):
+        return
+
+    @apihook('_initialize_narrow_environment', argc=0, conv=e_arch.CALL_CONV_CDECL)
+    def _initialize_narrow_environment(self, emu, argv, ctx={}):
+        return
+
+    @apihook('_configure_narrow_argv', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def _configure_narrow_argv(self, emu, argv, ctx={}):
+        return
+
+    @apihook('_set_fmode', argc=1, conv=e_arch.CALL_CONV_CDECL)
+    def _set_fmode(self, emu, argv, ctx={}):
+        return
+
+    @apihook('_itoa', argc=3, conv=e_arch.CALL_CONV_CDECL)
+    def _itoa(self, emu, argv, ctx={}):
+        return
+
+    @apihook('_itow', argc=3, conv=e_arch.CALL_CONV_CDECL)
+    def _itow(self, emu, argv, ctx={}):
+        return
